import jax
import jax.numpy as jnp
import jax.tree_util
from typing import Callable, Tuple, Dict


def sgd(
    learning_rate: float,
) -> Tuple[
    Callable[[Dict], Dict],
    Callable[[Dict, Dict, Dict], Tuple[Dict, Dict]]
]:
    """
    Stochastic Gradient Descent (SGD) optimizer.

    Args:
        learning_rate (float): The learning rate for the optimizer.

    Returns:
        Tuple[
            Callable[[Dict], Dict],
            Callable[[Dict, Dict, Dict], Tuple[Dict, Dict]]
        ]: A tuple containing the initialization and update functions for
        SGD.
    """

    def init(params: Dict) -> Dict:
        return params

    def update(params: Dict, grads: Dict, state: Dict) -> Tuple[Dict, Dict]:
        updated_params = jax.tree_util.tree_map(
            lambda p, g: p - learning_rate * g, params, grads
        )
        return updated_params, state

    return init, update


def adam(
    learning_rate: float,
    beta1: float = 0.9,
    beta2: float = 0.999,
    epsilon: float = 1e-8,
) -> Tuple[
    Callable[[Dict], Tuple[Dict, Dict]],
    Callable[[Dict, Dict, Tuple[Dict, Dict]], Tuple[Dict, Tuple[Dict, Dict]]],
]:
    """
    Adam optimizer.

    Args:
        learning_rate (float): The learning rate for the optimizer.
        beta1 (float): The exponential decay rate for the first moment
        estimates.
        beta2 (float): The exponential decay rate for the second moment
        estimates.
        epsilon (float): A small constant for numerical stability.

    Returns:
        Tuple[
            Callable[[Dict], Tuple[Dict, Dict]],
            Callable[
                [Dict, Dict, Tuple[Dict, Dict]],
                Tuple[Dict, Tuple[Dict, Dict]]
            ]
        ]: A tuple containing the initialization and update functions for Adam.
    """

    def init(params: Dict) -> Tuple[Dict, Dict]:
        m = jax.tree_util.tree_map(jnp.zeros_like, params)
        v = jax.tree_util.tree_map(jnp.zeros_like, params)
        return m, v

    def update(
        params: Dict, grads: Dict, state: Tuple[Dict, Dict]
    ) -> Tuple[Dict, Tuple[Dict, Dict]]:
        def update_m(m, g):
            return beta1 * m + (1 - beta1) * g

        def update_v(v, g):
            return beta2 * v + (1 - beta2) * jnp.square(g)

        m, v = state
        m = jax.tree_util.tree_map(update_m, m, grads)
        v = jax.tree_util.tree_map(update_v, v, grads)

        def m_hat_func(m):
            return m / (1 - beta1)

        def v_hat_func(v):
            return v / (1 - beta2)

        m_hat = jax.tree_util.tree_map(m_hat_func, m)
        v_hat = jax.tree_util.tree_map(v_hat_func, v)
        params = jax.tree_util.tree_map(
            lambda p, m, v: p - learning_rate * m / (jnp.sqrt(v) + epsilon),
            params,
            m_hat,
            v_hat,
        )
        return params, (m, v)

    return init, update


def rmsprop(
    learning_rate: float, decay: float = 0.9, epsilon: float = 1e-8
) -> Tuple[
    Callable[[Dict], Dict],
    Callable[[Dict, Dict, Dict], Tuple[Dict, Dict]]
]:
    """
    RMSprop optimizer.

    Args:
        learning_rate (float): The learning rate for the optimizer.
        decay (float): The decay rate for the moving average of squared
            gradients.
        epsilon (float): A small constant for numerical stability.

    Returns:
        Tuple[
            Callable[[Dict], Dict],
            Callable[[Dict, Dict, Dict], Tuple[Dict, Dict]]
        ]: A tuple containing the initialization and update functions for
        RMSprop.
    """

    def init(params: Dict) -> Dict:
        avg_sq_grad = jax.tree_util.tree_map(jnp.zeros_like, params)
        return avg_sq_grad

    def update(params: Dict, grads: Dict, state: Dict) -> Tuple[Dict, Dict]:
        avg_sq_grad = state
        avg_sq_grad = jax.tree_util.tree_map(
            lambda avg, g: decay * avg + (1 - decay) * jnp.square(g),
            avg_sq_grad, grads
        )
        params = jax.tree_util.tree_map(
            lambda p, avg, g: p - learning_rate * g / (
                jnp.sqrt(avg) + epsilon
            ),
            params,
            avg_sq_grad,
            grads,
        )
        return params, avg_sq_grad

    return init, update


def CustomOptimizer(
    learning_rate: float,
) -> Tuple[
    Callable[[Dict], Dict],
    Callable[[Dict, Dict, Dict], Tuple[Dict, Dict]]
]:
    """
    Custom optimizer example.

    Args:
        learning_rate (float): The learning rate for the optimizer.

    Returns:
        Tuple[
            Callable[[Dict], Dict],
            Callable[[Dict, Dict, Dict], Tuple[Dict, Dict]]
        ]: A tuple containing the initialization and update functions for
        the custom optimizer.
    """

    def init(params: Dict) -> Dict:
        return params

    def update(params: Dict, grads: Dict, state: Dict) -> Tuple[Dict, Dict]:
        # Example of a custom optimization algorithm
        updated_params = jax.tree_util.tree_map(
            lambda p, g: p - learning_rate * jnp.sin(g), params, grads
        )
        return updated_params, state

    return init, update



import jax
import jax.numpy as jnp
from jax.experimental import mesh_utils
from jax.sharding import PositionalSharding


def predict(params, inputs):
    for W, b in params:
        outputs = jnp.dot(inputs, W) + b
        inputs = jnp.maximum(outputs, 0)
    return outputs


def loss(params, batch):
    inputs, targets = batch
    predictions = predict(params, inputs)
    return jnp.mean(jnp.sum((predictions - targets) ** 2, axis=-1))


loss_jit = jax.jit(loss)
grad_fun = jax.jit(jax.grad(loss))


def init_layer(key, n_in, n_out):
    k1, k2 = jax.random.split(key)
    W = jax.random.normal(k1, (n_in, n_out)) / jnp.sqrt(n_in)
    b = jax.random.normal(k2, (n_out,))
    return W, b


def init_model(key, layer_sizes, batch_size):
    key, *keys = jax.random.split(key, len(layer_sizes))
    params = list(map(init_layer, keys, layer_sizes[:-1], layer_sizes[1:]))
    key, *keys = jax.random.split(key, 3)
    inputs = jax.random.normal(keys[0], (batch_size, layer_sizes[0]))
    targets = jax.random.normal(keys[1], (batch_size, layer_sizes[-1]))
    return params, (inputs, targets)


layer_sizes = [784, 8192, 8192, 8192, 10]
batch_size = 8192
params, batch = init_model(jax.random.PRNGKey(0), layer_sizes, batch_size)

# Define the device mesh and sharding
sharding = PositionalSharding(mesh_utils.create_device_mesh((8,)))

# Apply sharding to the batch and parameters
batch = jax.device_put(batch, sharding)
params = jax.device_put(params, sharding.replicate())

# Run the loss function with JIT compilation and parallelism
loss_value = loss_jit(params, batch)
print("Loss value:", loss_value)



from flax import linen as nn
import jax.numpy as jnp
from typing import Callable, Optional
from transformers import FlaxAutoModelForSeq2SeqLM, AutoTokenizer


class DenseLayer(nn.Module):
    features: int
    activation: Optional[Callable[[jnp.ndarray], jnp.ndarray]] = None

    def setup(self):
        self.dense = nn.Dense(features=self.features)

    def __call__(self, x: jnp.ndarray) -> jnp.ndarray:
        x = self.dense(x)
        if self.activation:
            x = self.activation(x)
        return x


class ConvolutionalLayer(nn.Module):
    features: int
    kernel_size: tuple
    strides: tuple = (1, 1)
    padding: str = "SAME"
    activation: Optional[Callable[[jnp.ndarray], jnp.ndarray]] = None

    def setup(self):
        self.conv = nn.Conv(
            features=self.features,
            kernel_size=self.kernel_size,
            strides=self.strides,
            padding=self.padding,
        )

    def __call__(self, x: jnp.ndarray) -> jnp.ndarray:
        x = self.conv(x)
        if self.activation:
            x = self.activation(x)
        return x


class TransformerLayer(nn.Module):
    """
    TransformerLayer integrates a pre-trained transformer model from the
    Hugging Face Transformers library.

    Attributes:
        model_name (str): The name of the pre-trained model to load.

    Methods:
        setup(): Initializes the tokenizer and model using the specified name.
        __call__(x: jnp.ndarray, max_length: int = 50) -> jnp.ndarray: Applies
            the transformer model to the input tensor.
    """
    model_name: str

    def setup(self):
        self.tokenizer = AutoTokenizer.from_pretrained(self.model_name)
        self.model = FlaxAutoModelForSeq2SeqLM.from_pretrained(self.model_name)

    def __call__(self, x: jnp.ndarray, max_length: int = 50) -> jnp.ndarray:
        if not isinstance(x, jnp.ndarray):
            raise ValueError("Input must be a jnp.ndarray")
        if not isinstance(max_length, int) or max_length <= 0:
            raise ValueError("max_length must be a positive integer")

        input_text = self.tokenizer.decode(x, skip_special_tokens=True)
        inputs = self.tokenizer(input_text, return_tensors="jax")
        outputs = self.model.generate(**inputs, max_length=max_length)
        output_text = self.tokenizer.decode(outputs[0],
                                            skip_special_tokens=True)
        return self.tokenizer(output_text, return_tensors="jax")["input_ids"]



# Add a new comment to ensure changes are detected for commit
# Trigger CI/CD workflow - 2024-07-13
# This comment is added to trigger a new workflow run - 2023-05-11
# This comment is added to trigger another workflow run - 2024-07-13
import os
from transformers import AutoModelForSeq2SeqLM, AutoTokenizer
import torch


class TransformerModel:
    def __init__(self, model_name: str, device: str = "cpu"):
        self.model_name = model_name
        self.device = torch.device(device)
        self.tokenizer = AutoTokenizer.from_pretrained(model_name)
        self.model = AutoModelForSeq2SeqLM.from_pretrained(model_name).to(
            self.device
        )

    def save_model(self, save_directory: str):
        os.makedirs(save_directory, exist_ok=True)
        self.tokenizer.save_pretrained(
            os.path.join(save_directory, "tokenizer")
        )
        self.model.save_pretrained(os.path.join(save_directory, "model"))

    def load_model(self, load_directory: str):
        try:
            self.tokenizer = AutoTokenizer.from_pretrained(
                os.path.join(load_directory, "tokenizer")
            )
            self.model = AutoModelForSeq2SeqLM.from_pretrained(
                os.path.join(load_directory, "model")
            ).to(self.device)
        except Exception as e:
            print(f"Error loading model or tokenizer: {e}")

    def generate_text(self, input_text: str, max_length: int = 50) -> str:
        if not isinstance(input_text, str):
            raise ValueError("input_text must be a string")
        if not isinstance(max_length, int) or max_length <= 0:
            raise ValueError("max_length must be a positive integer")

        inputs = self.tokenizer(input_text, return_tensors="pt").to(
            self.device)
        outputs = self.model.generate(
            inputs["input_ids"],
            max_length=max_length
        )
        return self.tokenizer.decode(outputs[0], skip_special_tokens=True)



import jax
import jax.numpy as jnp
from flax import linen as nn
from typing import Callable


class VocabParallelEmbedding(nn.Module):
    num_embeddings: int
    embedding_dim: int
    init_method: Callable = jax.nn.initializers.normal()

    def setup(self):
        self.embedding = self.param(
            'embedding',
            self.init_method,
            (self.num_embeddings, self.embedding_dim)
        )

    def __call__(self, input_ids: jnp.ndarray) -> jnp.ndarray:
        # Split the embedding matrix across devices
        num_devices = jax.local_device_count()
        embedding_split = jnp.array_split(self.embedding, num_devices, axis=0)

        # Define a function to perform the embedding lookup on each device
        def lookup_on_device(embedding, input_ids):
            return embedding[input_ids]

        # Use pmap to parallelize the embedding lookup across devices
        parallel_lookup = jax.pmap(
            lookup_on_device,
            in_axes=(0, None),
            out_axes=0
        )
        result = parallel_lookup(embedding_split, input_ids)

        # Gather the results from all devices
        return jnp.concatenate(result, axis=0)



import jax.numpy as jnp

__all__ = ['relu', 'sigmoid', 'tanh', 'leaky_relu', 'CustomActivation']


def relu(x: jnp.ndarray) -> jnp.ndarray:
    return jnp.maximum(0, x)


def sigmoid(x: jnp.ndarray) -> jnp.ndarray:
    return 1 / (1 + jnp.exp(-x))


def tanh(x: jnp.ndarray) -> jnp.ndarray:
    return jnp.tanh(x)


def leaky_relu(x: jnp.ndarray, negative_slope: float = 0.01) -> jnp.ndarray:
    return jnp.where(x > 0, x, negative_slope * x)


class CustomActivation:
    @staticmethod
    def forward(x: jnp.ndarray) -> jnp.ndarray:
        # Example of a custom activation function
        return jnp.sin(x)


# Import main components of the nextgenjax package
from .model import NextGenModel
from .layers import DenseLayer, ConvolutionalLayer
from .transformer_models import TransformerModel
from .custom_layers import CustomLayer
from .optimizers import CustomOptimizer
from .activations import CustomActivation

__all__ = ['NextGenModel', 'DenseLayer', 'ConvolutionalLayer', 'TransformerModel', 'CustomLayer', 'CustomOptimizer', 'CustomActivation']


import haiku as hk
import jax
import jax.numpy as jnp
import optax

class NextGenModel(hk.Module):
    def __init__(self, num_layers, hidden_size, num_heads, dropout_rate):
        super().__init__()
        assert hidden_size % num_heads == 0, "hidden_size must be divisible by num_heads"
        assert num_layers > 0, "num_layers must be greater than 0"
        assert 0 <= dropout_rate < 1, "dropout_rate must be between 0 and 1"
        self.num_layers = num_layers
        self.hidden_size = hidden_size
        self.num_heads = num_heads
        self.dropout_rate = dropout_rate

    def __call__(self, x, train: bool = False):
        for _ in range(self.num_layers):
            x = self.encoder_layer(x, train)
        return x

    def encoder_layer(self, x, train: bool):
        # Self-attention
        residual = x
        x = hk.LayerNorm(axis=-1, create_scale=True, create_offset=True, param_axis=(), scale_init=hk.initializers.Constant(1.0))(x)
        x = hk.MultiHeadAttention(
            num_heads=self.num_heads,
            key_size=self.hidden_size // self.num_heads,
            model_size=self.hidden_size,
            w_init=hk.initializers.VarianceScaling(2.0)
        )(x, x, x)
        x = hk.dropout(hk.next_rng_key(), self.dropout_rate, x) if train else x
        x = x + residual

        # Feed-forward
        residual = x
        x = hk.LayerNorm(axis=-1, create_scale=True, create_offset=True, param_axis=(), scale_init=hk.initializers.Constant(1.0))(x)
        x = hk.Linear(output_size=self.hidden_size * 4)(x)
        x = jax.nn.gelu(x)
        x = hk.Linear(output_size=self.hidden_size)(x)
        x = hk.dropout(hk.next_rng_key(), self.dropout_rate, x) if train else x
        x = x + residual

        return x

def create_model(num_layers, hidden_size, num_heads, dropout_rate):
    def _model(x, train=False):
        model = NextGenModel(num_layers, hidden_size, num_heads, dropout_rate)
        return model(x, train)
    return hk.transform(_model)

def create_optimizer(learning_rate: float = 1e-3):
    return optax.adam(learning_rate)


# Updated to trigger CI/CD workflow - 2023-05-10
# Triggering a new CI/CD workflow run to verify fixes
import jax
import jax.numpy as jnp
import jax.tree_util
import jax.debug
from jax import value_and_grad
from flax.training import train_state
from typing import Any, Callable, Dict, Tuple, Union
from .model import NextGenModel
import optax
import haiku as hk

# Type alias for optimizer
OptimizerType = optax.GradientTransformation

def create_model(num_layers, hidden_size, num_heads, dropout_rate):
    def _model(x, train=False):
        model = NextGenModel(num_layers, hidden_size, num_heads, dropout_rate)
        return model(x, train)
    return hk.transform(_model)


def create_train_state(
    rng: jax.random.PRNGKey,
    model: Any,
    optimizer: OptimizerType,
    hidden_size: int,
    sequence_length: int = 64,
) -> train_state.TrainState:
    """
    Creates initial training state.

    Args:
        rng (jax.random.PRNGKey): The random number generator key.
        model (Any): The model to be trained (Haiku transformed or regular module).
        optimizer (OptimizerType): The optimizer to use.
        hidden_size (int): The hidden size of the model.
        sequence_length (int): The sequence length for the dummy input. Default is 64.

    Returns:
        train_state.TrainState: The initial training state.

    Raises:
        TypeError: If the model is neither a Haiku transformed function nor a regular Haiku module.
    """
    dummy_input = jnp.ones([1, sequence_length, hidden_size])

    if isinstance(model, hk.Transformed):
        params = model.init(rng, dummy_input)
        apply_fn = lambda params, rng, *args, **kwargs: model.apply(params, rng, *args, **kwargs)
    elif isinstance(model, hk.Module):
        params = model.init(rng, dummy_input)["params"]
        apply_fn = lambda params, rng, *args, **kwargs: model.apply({"params": params}, rng, *args, **kwargs)
    else:
        raise TypeError("Model must be either a Haiku transformed function or a regular Haiku module")

    # Add print statement to check apply_fn output
    print(f"apply_fn output type: {type(apply_fn(params, rng, dummy_input))}")

    # Add print statement to check optimizer type
    print(f"Optimizer type: {type(optimizer)}")

    return train_state.TrainState.create(
        apply_fn=apply_fn,
        params=params,
        tx=optimizer,
    )

def train_step(
    state: train_state.TrainState,
    batch: Dict[str, jnp.ndarray],
    loss_fn: Callable[[jnp.ndarray, Callable, Dict[str, jnp.ndarray], jax.random.PRNGKey], float],
    rng: jax.random.PRNGKey,
) -> Tuple[train_state.TrainState, Dict[str, float], jax.random.PRNGKey]:
    """
    Performs a single training step.

    Args:
        state (train_state.TrainState): The current training state.
        batch (Dict[str, jnp.ndarray]): A batch of training data.
        loss_fn (Callable[[jnp.ndarray, Callable, Dict[str, jnp.ndarray], jax.random.PRNGKey], float]): A function to
        compute the loss given the model parameters, apply function, batch, and RNG key.
        rng (jax.random.PRNGKey): The random number generator key.

    Returns:
        Tuple[train_state.TrainState, Dict[str, float], jax.random.PRNGKey]:
        The updated training state, metrics, and new RNG key.
    """
    print("train_step: Input types:")
    print(f"  state: {type(state)}")
    print(f"  batch: {type(batch)}")
    print(f"  loss_fn: {type(loss_fn)}")
    print(f"  rng: {type(rng)}")

    def loss_and_grad(params):
        print(f"loss_and_grad: params type: {type(params)}")
        print(f"loss_and_grad: state.apply_fn type: {type(state.apply_fn)}")
        print(f"loss_and_grad: batch type: {type(batch)}")
        print(f"loss_and_grad: rng type: {type(rng)}")
        loss = loss_fn(params, state.apply_fn, batch, rng)
        print(f"loss_and_grad: loss type: {type(loss)}")
        print(f"loss_and_grad: loss value: {loss}")
        return loss

    grad_fn = jax.value_and_grad(loss_and_grad)
    loss, grads = grad_fn(state.params)
    print(f"train_step: grads type: {type(grads)}")
    state = state.apply_gradients(grads=grads)
    metrics = {"loss": float(loss)}  # Convert loss to float
    return state, metrics, rng


def train_model(
    model_params: Tuple[int, int, float],
    train_dataset: Any,
    num_epochs: int,
    optimizer: OptimizerType,
    loss_fn: Callable[[jnp.ndarray, Callable, Dict[str, jnp.ndarray], jax.random.PRNGKey], float],
    hidden_size: int,
    sequence_length: int,
    rng: jax.random.PRNGKey,
) -> Tuple[train_state.TrainState, Dict[str, float]]:
    """
    Trains the model.

    Args:
        model_params (Tuple[int, int, float]): Parameters for creating the model (num_layers, num_heads, dropout_rate).
        train_dataset (Any): The training dataset.
        num_epochs (int): The number of epochs to train for.
        optimizer (OptimizerType): The optimizer to use.
        loss_fn (Callable[[jnp.ndarray, Callable, Dict[str, jnp.ndarray], jax.random.PRNGKey], float]): A function to
        compute the loss given the model parameters, apply function, batch, and RNG key.
        hidden_size (int): The hidden size of the model.
        sequence_length (int): The sequence length for the input.
        rng (jax.random.PRNGKey): The random number generator key.

    Returns:
        Tuple[train_state.TrainState, Dict[str, float]]: The final training
        state and metrics.
    """
    print("train_model: Input types:")
    print(f"  model_params: {type(model_params)}")
    print(f"  train_dataset: {type(train_dataset)}")
    print(f"  num_epochs: {type(num_epochs)}")
    print(f"  optimizer: {type(optimizer)}")
    print(f"  loss_fn: {type(loss_fn)}")
    print(f"  hidden_size: {type(hidden_size)}")
    print(f"  sequence_length: {type(sequence_length)}")
    print(f"  rng: {type(rng)}")

    def data_loader(dataset):
        for batch in dataset:
            yield batch

    def debug_loss_fn(params, apply_fn, batch, rng):
        print(f"Debug loss_fn - Params type: {type(params)}")
        print(f"Debug loss_fn - Apply_fn type: {type(apply_fn)}")
        print(f"Debug loss_fn - Batch type: {type(batch)}")
        print(f"Debug loss_fn - RNG type: {type(rng)}")
        loss = loss_fn(params, apply_fn, batch, rng)
        print(f"Debug loss_fn - Loss type: {type(loss)}")
        print(f"Debug loss_fn - Loss value: {loss}")
        return loss

    if len(model_params) != 3:
        raise ValueError("model_params must contain exactly 3 elements: num_layers, num_heads, and dropout_rate")

    model = create_model(model_params[0], hidden_size, model_params[1], model_params[2])
    rng, init_rng = jax.random.split(rng)
    state = create_train_state(init_rng, model, optimizer, hidden_size, sequence_length)

    metrics_history = []
    for epoch in range(num_epochs):
        epoch_loss = []
        for batch in data_loader(train_dataset):
            rng, step_rng = jax.random.split(rng)
            state, metrics, rng = train_step(state, batch, debug_loss_fn, step_rng)
            epoch_loss.append(metrics["loss"])
        avg_loss = float(jnp.mean(jnp.array(epoch_loss)))  # Convert to float
        metrics_history.append({"loss": avg_loss})
        print(f"Epoch {epoch + 1}, Loss: {avg_loss}")

    return state, metrics_history


from setuptools import setup, find_packages

setup(
    name="nextgenjax",
    version="0.1.1",
    author="kasinadhsarma",
    author_email="kasinadhsarma@gmail.com",
    description="A JAX-based neural network library surpassing Google DeepMind's Haiku and Optax",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    url="https://github.com/VishwamAI/NextGenJAX",
    packages=find_packages(where="src"),
    package_dir={"": "src"},
    install_requires=[
        "jax==0.4.30",
        "jaxlib==0.4.30",
        "flax==0.8.5",
        "optax==0.2.3",
        "numpy>=1.26.4",
        "scipy>=1.10.1",
        "fairscale>=0.4.6",
        "transformers>=4.30.2",
        "chex>=0.1.86",
        "dm-haiku>=0.0.5",
    ],
    extras_require={
        "dev": [
            "pytest==6.2.4",
            "flake8==3.9.2",
        ],
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.9",
)



import jax.numpy as jnp
from flax import linen as nn
from typing import Callable, Optional


class CustomLayer(nn.Module):
    """
    A custom layer that includes a dense layer and an optional activation
    function.

    Attributes:
        features (int): The number of output features for the dense layer.
        activation (Optional[Callable[[jnp.ndarray], jnp.ndarray]]): An
        optional activation function to apply after the dense layer.
    """

    features: int
    activation: Optional[Callable[[jnp.ndarray], jnp.ndarray]] = None

    def setup(self):
        """Sets up the dense layer with the specified number of features."""
        self.dense = nn.Dense(features=self.features)

    def __call__(self, x: jnp.ndarray) -> jnp.ndarray:
        """
        Applies the dense layer and the optional activation function to the
        input.

        Args:
            x (jnp.ndarray): The input array.

        Returns:
            jnp.ndarray: The output array after applying the dense layer and
            the optional activation function.
        """
        x = self.dense(x)
        if self.activation:
            x = self.activation(x)
        return x
